"""Configuration variables.
"""

from collections import OrderedDict as ODict
import pathlib


# Database location.
here = pathlib.Path(__file__)
DB_PATH = here.parent / 'instance' / 'queue.sqlite'

# Database schema.
# NOTE: Changing this configuration will not change the database schema.
# In order to do so, the database must be re-initialized.
QUEUE_TABLENAME = 'queue'
WORK_TABLENAME = 'work'
SCHEMA = ODict([
    (QUEUE_TABLENAME, ODict(
        task_id='TEXT PRIMARY KEY NOT NULL',
        queue_name='TEXT',
        position='INTEGER UNIQUE NOT NULL',
        published='TIMESTAMP NOT NULL',
        args='BLOB',
        kwargs='BLOB',
    )),
    (WORK_TABLENAME, ODict(
        task_id='TEXT PRIMARY KEY NOT NULL',
        queue_name='TEXT',
        started='TIMESTAMP NOT NULL',
        status='TEXT',
        exc_type='BLOB',
        exc_value='BLOB',
        exc_traceback='BLOB',
    )),
])



"""Useful commands.

Classes:
QueueItem
WorkItem

Functions:
clear_queue - Clear the queue.
clear_work - Clear the table of work.
task_items - Info about items in the queue.
work_items - Info about items in the table of work.
"""

from collections import namedtuple
import pickle
import sqlite3
import traceback

from .config import DB_PATH, QUEUE_TABLENAME, SCHEMA, WORK_TABLENAME
from .utils import echo


QueueItem = namedtuple('QueueItem', SCHEMA[QUEUE_TABLENAME].keys())
WorkItem = namedtuple('WorkItem', SCHEMA[WORK_TABLENAME].keys())


def clear_queue():
    """Clear the task queue."""
    con = sqlite3.connect(str(DB_PATH))
    with con:
        con.execute(
            f"""
            DELETE FROM {QUEUE_TABLENAME}
            """
        )
    con.close()
    echo("Cleared the queue.")


def clear_work():
    """Clear the table of work."""
    con = sqlite3.connect(str(DB_PATH))
    with con:
        con.execute(
            f"""
            DELETE FROM {WORK_TABLENAME}
            """
        )
    con.close()
    echo("Cleared table of work.")


def task_items(max_entries=None):
    """Information about the items in the task queue. Returns a
    generator of QueueItems.

    Keyword arguments:
    max_entries - (int) (Default: None) Maximum number of items to
        return. Default is to return all entries.
    """
    con = sqlite3.connect(str(DB_PATH))
    with con:
        c = con.execute(
            f"""
            SELECT task_id, queue_name, position, published, args, kwargs
            FROM {QUEUE_TABLENAME}
            ORDER BY position DESC
            """
        )
    if max_entries:
        items = map(QueueItem._make, c.fetchmany(size=max_entries))
    else:
        items = map(QueueItem._make, c.fetchall())
    con.close()
    return items


def work_items(max_entries=None):
    """Information about the items in the work queue. Returns a
    generator of WorkItems.

    Keyword arguments:
    max_entries - (int) (Default: None) Maximum number of items to
        return. Default is to return all entries.
    """
    con = sqlite3.connect(str(DB_PATH))
    with con:
        c = con.execute(
            f"""
            SELECT
                task_id, queue_name, started, status,
                exc_type, exc_value, exc_traceback
            FROM {WORK_TABLENAME}
            ORDER BY started ASC
            """
        )
    if max_entries:
        entries = c.fetchmany(size=max_entries)
    else:
        entries = c.fetchall()
    con.close()
    for row in entries:
        exc_type, exc_value = tuple(pickle.loads(bin_) for bin_ in row[-3:-1])
        exc_tb = ''.join(pickle.loads(row[-1]).format())
        yield WorkItem(*row[:-3], exc_type, exc_value, exc_tb)



"""Utility functionalities.

Functions:
echo
"""

import functools

try:
    import click
    feedback_func = click.echo
except ModuleNotFoundError:
    feedback_func = print


@functools.wraps(feedback_func)
def echo(*args, **kwargs):
    """Function for giving feedback. Tries to use click.echo, but
    defaults to print if Click is not installed.
    """
    return feedback_func(*args, **kwargs)



"""Implements worker classes.

Classes:
Worker
"""

import pickle
import sys
import traceback

from .utils import echo


class Worker:
    """Basic worker class.

    Example Usage:
        task = AdderTask()
        worker = Worker(task=task)
        worker.start()
    """

    def __init__(self, task):
        self.task = task
        self.waiting = False

    def start(self):
        """Begin working on the assigned type of task."""
        while True:
            try:
                # Read database.
                dequeued_item = self.task.dequeue()
                if not dequeued_item:
                    if not self.waiting:
                        echo(f"Waiting for next task... (Ctrl + C to quit)\n")
                        self.waiting = True
                    continue
                self.waiting = False
                self.task.set_status('dequeued')
                task_id, _, _, _, task_args, task_kwargs = dequeued_item
                task_args = pickle.loads(task_args)
                task_kwargs = pickle.loads(task_kwargs)
                # Run.
                exc_info = None
                echo(f"Running task: {task_id}")
                self.task.set_status('running')
                try:
                    self.task.run(*task_args, **task_kwargs)
                except:
                    # Any Exceptions will be saved.
                    exc_info = list(sys.exc_info())
                    exc_info[-1] = traceback.extract_tb(exc_info[-1])
                if exc_info:
                    echo(f"Error in task: {task_id}\n")
                    self.task.set_status('error')
                    self.task.record_exc(*exc_info)
                else:
                    echo(f"Finished task: {task_id}\n")
                    self.task.set_status('complete')
            except KeyboardInterrupt:
                self.quit()
                break

    def quit(self):
        """Stop working."""
        echo("Quitting")



"""Implements broker classes.

Classes:
Broker
"""

import datetime as dt
import pickle
import sqlite3

from .config import DB_PATH, QUEUE_TABLENAME, WORK_TABLENAME


class Broker:
    """Implements a basic FIFO queue using SQLite.
    """

    def enqueue(self, item, queue_name):
        """Enqueue the given item in the queue with the given name."""
        con = sqlite3.connect(str(DB_PATH))
        with con:
            c = con.execute(
                f"""
                SELECT position
                FROM {QUEUE_TABLENAME}
                ORDER BY position DESC
                """
            )
            max_position = c.fetchone()  # Returns a tuple.
            new_position = max_position[0] + 1 if max_position else 0
            con.execute(
                f"""
                INSERT INTO {QUEUE_TABLENAME} (
                    task_id,
                    queue_name,
                    position,
                    published,
                    args,
                    kwargs
                )
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (
                    item['task_id'], queue_name, new_position,
                    dt.datetime.now(), item['args'], item['kwargs'],
                ),
            )
        con.close()

    def dequeue(self, queue_name):
        """Dequeue the next item (item with lowest position number) from
        the queue with the given name and return it.
        """
        con = sqlite3.connect(str(DB_PATH))
        con.row_factory = sqlite3.Row
        with con:
            c = con.execute(
                f"""
                SELECT
                    task_id,
                    queue_name,
                    position,
                    published,
                    args,
                    kwargs
                FROM {QUEUE_TABLENAME}
                WHERE queue_name = ?
                ORDER BY position ASC
                """,
                (queue_name,),
            )
            dequeued_item = c.fetchone()
            if dequeued_item:
                item_id = dequeued_item['task_id']
                con.execute(
                    f"""
                    DELETE FROM {QUEUE_TABLENAME}
                    WHERE task_id = ?
                    """,
                    (item_id,),
                )
        con.close()
        return dequeued_item

    def add_work(self, item, status):
        """Add the given item to the work table."""
        con = sqlite3.connect(str(DB_PATH))
        with con:
            con.execute(
                f"""
                INSERT INTO {WORK_TABLENAME} (
                    task_id,
                    queue_name,
                    started,
                    status
                )
                VALUES (?, ?, ?, ?)
                """,
                (item.task_id, item.task_name, dt.datetime.now(), status),
            )
        con.close()

    def remove_work(self, item):
        """Remove the given item from the work table."""
        con = sqlite3.connect(str(DB_PATH))
        with con:
            con.execute(
                f"""
                DELETE FROM {WORK_TABLENAME}
                WHERE task_id = ?
                """,
                (item.task_id,),
            )
        con.close()

    def update_status(self, item, status):
        """Update the status of the given item in the work table."""
        con = sqlite3.connect(str(DB_PATH))
        with con:
            con.execute(
                f"""
                UPDATE {WORK_TABLENAME}
                SET status = ?
                WHERE task_id = ?
                """,
                (status, item.task_id),
            )
        con.close()

    def update_exc_info(self, item, type_, value, traceback):
        """Update exception info for the given item."""
        con = sqlite3.connect(str(DB_PATH))
        # Pickle objects to be stored as BLOB.
        ptype = pickle.dumps(type_, protocol=pickle.HIGHEST_PROTOCOL)
        pvalue = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL)
        ptraceback = pickle.dumps(traceback, protocol=pickle.HIGHEST_PROTOCOL)
        with con:
            con.execute(
                f"""
                UPDATE {WORK_TABLENAME}
                SET
                    exc_type = ?,
                    exc_value = ?,
                    exc_traceback = ?
                WHERE task_id = ?
                """,
                (ptype, pvalue, ptraceback, item.task_id),
            )
        con.close()



"""Implements task classes.

Classes:
BaseTask - Base task class.
"""

import abc
import pickle
import uuid

from .broker import Broker
from .utils import echo


class BaseTask(abc.ABC):
    """Base task class.

    Example Usage:
        class AdderTask(BaseTask):
            task_name = 'AdderTask'
            def run(self, a, b):
                result = a + b
                return result

        adder = AdderTask()
        adder.delay(9, 34)
    """

    # Status options are:
    #   detached - Instantiated, not queued.
    #   enqueued - Queued via Broker.
    #   dequeued - Dequeued via Worker.
    #   running - Running via Worker.
    #   error - Exception encountered during run.
    #   complete - Run complete.
    _status_options = (
        'detached', 'enqueued', 'dequeued', 'running', 'error',
        'complete')

    def __init__(self):
        self.broker = Broker()
        self.set_status('detached')

    @property
    @classmethod
    @abc.abstractmethod
    def task_name(cls):
        pass

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        """Subclasses must implement their business logic here."""
        pass

    def delay(self, *args, **kwargs):
        """Have the broker enqueue this task, thereby delaying its
        execution until some future time.
        """
        try:
            self.task_id = str(uuid.uuid4())
            task = dict(
                task_id=self.task_id,
                args=pickle.dumps(args, protocol=pickle.HIGHEST_PROTOCOL),
                kwargs=pickle.dumps(kwargs, protocol=pickle.HIGHEST_PROTOCOL),
            )
            self.broker.enqueue(
                item=task, queue_name=self.task_name)
            self.set_status('enqueued')
            echo(f"Queued task: {self.task_id}")
        except Exception:
            raise RuntimeError(
                f"Unable to publish task {self.task_id} to the broker.")

    def dequeue(self):
        """Dequeue one task of this type."""
        item = self.broker.dequeue(queue_name=self.task_name)
        try:
            self.task_id = item['task_id']
        except TypeError:
            self.task_id = None
        return item

    def set_status(self, status):
        """Set status of the BaseTask instance."""
        if status not in self._status_options:
            raise ValueError(f"Task status {status!r} not recognized.")
        self.status = status
        if status == 'running':
            self.broker.add_work(self, status)
        elif status == 'complete':
            self.broker.update_status(self, status)
            self.broker.remove_work(self)

    def record_exc(self, type_, value, traceback):
        """Record the given traceback information."""
        self.broker.update_exc_info(self, type_, value, traceback)

    def __repr__(self):
        return (f"<{self.__class__.__name__}(status={self.status})>")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['soonq']

package_data = \
{'': ['*'], 'soonq': ['instance/*']}

setup_kwargs = {
    'name': 'soonq',
    'version': '0.3.0',
    'description': 'Subprocess-based task queue.',
    'long_description': '# SoonQ\nA subprocess-based task queue.\n\n## Introduction\nSoonQ implements a simple FIFO queue using SQLite. It was created primarily for running long simulations.\n\nAs of yet, the vision of a subprocess-based workflow has not been realized. However, the package still works as a task queue.\n\n## Installation\n`pip install soonq`\n\n## Usage\nUsers must create their own subclass of `soonq.BaseTask`. Subclasses must define a `run()` method, which contains the business logic for the task (what we care about). Input arguments to this method are restricted to being serializable via the pickle module.\n\n## Running the examples\n\nExample files are included in the examples directory. Clone SoonQ in your desired location...\n\n`C:\\desired\\location>git clone https://github.com/n8jhj/SoonQ.git`\n\n...and then navigate into the `SoonQ` directory and install it. Be careful to include the dot!\n\n`pip install .`\n\nNow run the following in two separate terminals:\n\n**Terminal 1:**\n\nRun the same script a couple times.\n\n    C:\\...\\SoonQ>python -m examples timer_task\n    Queued task: 913d56e9-a609-4b84-b937-479a94716527\n\n    C:\\...\\SoonQ>python -m examples timer_task\n    Queued task: da952424-98d9-42e1-8851-91a30924b94b\n\n    C:\\...\\SoonQ>\n\n**Terminal 2:**\n\n    C:\\...\\SoonQ>python -m examples timer_worker\n    Running task: 913d56e9-a609-4b84-b937-479a94716527\n    1/3 Sleeping 3 seconds...\n    2/3 Sleeping 3 seconds...\n    3/3 Sleeping 3 seconds...\n    Slept 9 seconds total.\n    Finished task: 913d56e9-a609-4b84-b937-479a94716527\n\n    Running task: da952424-98d9-42e1-8851-91a30924b94b\n    1/3 Sleeping 3 seconds...\n    2/3 Sleeping 3 seconds...\n    3/3 Sleeping 3 seconds...\n    Slept 9 seconds total.\n    Finished task: da952424-98d9-42e1-8851-91a30924b94b\n\n    Waiting for next task... (Ctrl + C to quit)\n\nWith the worker running, more tasks can be enqueued and will be immediately processed.\n\nPress `Ctrl + C` to quit the worker.\n\n    Quitting\n\n    C:\\...\\SoonQ>\n\n## Etymology\nThis project is named after my friend Soon-Kyoo, with whom I enjoyed countless bouts of epic ping-pong in college. People call him Q, for short.\n',
    'author': 'Nathaniel Jones',
    'author_email': 'nathaniel.j.jones@wsu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/n8jhj/SoonQ',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
